package de.hdm_stuttgart.mi.sd1.test.ignore_me;

import org.junit.Assert;

import java.lang.reflect.*;
import java.util.Arrays;


public class Wrapper<T> {

  static private void assertFail(final String msg, final Exception e) {
    final String exceptionMsg = e.getMessage();
    if (null == exceptionMsg) {
      Assert.fail(msg + ":\n  " + e.getCause());
    } else {
      Assert.fail(msg + ":\n  " + exceptionMsg);
    }
  }

  static public void assertFinal(final Class classZ, final String fieldName) {
    try {
      final int modifier = classZ.getDeclaredField(fieldName).getModifiers();
      Assert.assertTrue("»" + fieldName + "« can be modified, no immutable!", Modifier.isFinal(modifier));

    } catch (NoSuchFieldException e) {
      Assert.fail("No such field »" + fieldName + "«");
    }
  }

  private final Class<T> classz;
  private T instance;

  public Wrapper(final Class<T> classz, final P... parameter) {
    this.classz = classz;
    try {

      final Constructor<T> constructor = classz.getDeclaredConstructor(
          Arrays.stream(parameter).map(P::getType).toArray(Class[]::new)
      );
      instance = constructor.newInstance(
          Arrays.stream(parameter).map(P::getParameter).toArray()
      );
    } catch (final NoSuchMethodException e) {
      assertFail("No suitable constructor found in class " + classz.getName(), e);
    } catch (final InstantiationException | IllegalAccessException | InvocationTargetException e) {
      assertFail("Unable to instantiate " + classz.getName(), e);
    }
  }

  public T getInstance() {
    return instance;
  }
  private P get(final String name) {
    try {
      final Field field = classz.getDeclaredField(name);
      return new P(field.getType(), field.get(instance));
    } catch (final NoSuchFieldException e) {
      assertFail("No such field '" + name + " in " + classz.getName(), e);

    } catch (IllegalAccessException e) {
      assertFail("Failure accessing field '" + name , e);
    }
    return null;
  }

  private P invoke(final String name, final P... parameter) {
    try {
      final Method method =
          classz.getDeclaredMethod(name, Arrays.stream(parameter).map(P::getType).toArray(Class[]::new));

      return new P(
      method.getReturnType(),
      method.invoke(instance, Arrays.stream(parameter).map(P::getParameter).toArray())
      );
    } catch (final NoSuchMethodException e) {
      assertFail("No such method found", e);

    } catch (IllegalAccessException | InvocationTargetException e) {
      assertFail("Failure calling method '" + name , e);
    }
    return null;
  }

  // Field access

  /**
   * Field access convenience method.
   *
   * @param name Field name.
   * @return Field value.
   */
  public byte getByte(final String name) {
    return get(name).getByte();
  }
  /**
   * Field access convenience method.
   *
   * @param name Field name.
   * @return Field value.
   */
  public short getShort(final String name) {
    return get(name).getShort();
  }
  /**
   * Field access convenience method.
   *
   * @param name Field name.
   * @return Field value.
   */
  public int getInt(final String name) {
    return get(name).getInt();
  }
  /**
   * Field access convenience method.
   *
   * @param name Field name.
   * @return Field value.
   */
  public long getLong(final String name) {
    return get(name).getLong();
  }
  /**
   * Field access convenience method.
   *
   * @param name Field name.
   * @return Field value.
   */
  public char getChar(final String name) {
    return get(name).getChar();
  }
  /**
   * Field access convenience method.
   *
   * @param name Field name.
   * @return Field value.
   */
  public float getFloat(final String name) {
    return get(name).getFloat();
  }
  /**
   * Field access convenience method.
   *
   * @param name Field name.
   * @return Field value.
   */
  public double getDouble(final String name) {
    return get(name).getDouble();
  }
  /**
   * Field access convenience method.
   *
   * @param name Field name.
   * @return Field value.
   */
  public boolean getBoolean(final String name) {
    return get(name).getBoolean();
  }
  /**
   * Field access convenience method.
   *
   * @param name Field name.
   * @return Field value.
   */
  public Object getObject(final String name) {
    return get(name).get();
  }

  // Invoke methods

  /**
   * Named method call.
   *
   * @param name Callee's name.
   * @param parameter Call parameter
   * @return Callee's return value
   */
  public byte invokeByte(final String name, final P... parameter) {
    return invoke(name, parameter).getByte();
  }
  /**
   * Named method call.
   *
   * @param name Callee's name.
   * @param parameter Call parameter
   * @return Callee's return value
   */
  public short invokeShort(final String name, final P... parameter) {
    return invoke(name, parameter).getShort();
  }
  /**
   * Named method call.
   *
   * @param name Callee's name.
   * @param parameter Call parameter
   * @return Callee's return value
   */
  public int invokeInt(final String name, final P... parameter) {
    return invoke(name, parameter).getInt();
  }
  /**
   * Named method call.
   *
   * @param name Callee's name.
   * @param parameter Call parameter
   * @return Callee's return value
   */
  public long invokeLong(final String name, final P... parameter) {
    return invoke(name, parameter).getLong();
  }
  /**
   * Named method call.
   *
   * @param name Callee's name.
   * @param parameter Call parameter
   * @return Callee's return value
   */
  public char invokeChar(final String name, final P... parameter) {
    return invoke(name, parameter).getChar();
  }
  /**
   * Named method call.
   *
   * @param name Callee's name.
   * @param parameter Call parameter
   * @return Callee's return value
   */
  public float invokeFloat(final String name, final P... parameter) {
    return invoke(name, parameter).getFloat();
  }
  /**
   * Named method call.
   *
   * @param name Callee's name.
   * @param parameter Call parameter
   * @return Callee's return value
   */
  public double invokeDouble(final String name, final P... parameter) {
    return invoke(name, parameter).getDouble();
  }
  /**
   * Named method call.
   *
   * @param name Callee's name.
   * @param parameter Call parameter
   * @return Callee's return value
   */
  public boolean invokeBoolean(final String name, final P... parameter) {
    return invoke(name, parameter).getBoolean();
  }
  /**
   * Named method call.
   *
   * @param name Callee's name.
   * @param parameter Call parameter
   * @return Callee's return value
   */
  public Object invokeObject(final String name, final P... parameter) {
    return invoke(name, parameter).get();
  }
}