package de.hdm_stuttgart.mi.sd1.test.task2;

import de.hdm_stuttgart.mi.exam.unitmarking.ExaminationTestDefaults;
import de.hdm_stuttgart.mi.exam.unitmarking.Marking;
import de.hdm_stuttgart.mi.sd1.task2.Direction;
import de.hdm_stuttgart.mi.sd1.task2.Position;
import de.hdm_stuttgart.mi.sd1.test.ignore_me.P;
import de.hdm_stuttgart.mi.sd1.test.ignore_me.Wrapper;
import org.junit.Assert;
import org.junit.FixMethodOrder;
import org.junit.Test;
import org.junit.runners.MethodSorters;

@FixMethodOrder(MethodSorters.NAME_ASCENDING)
@SuppressWarnings({"UnusedDeclaration"})
public class TestPosition extends ExaminationTestDefaults {

  @Test @Marking(points = 6) public void test_100_constructor() {
    assertCreatePosition(3, 2);
    assertCreatePosition(10, 9);
    assertCreatePosition(5, 5);
    assertCreatePosition(0, 7);
  }

  @Test @Marking(points = 6) public void test_140_legalMoves() {
    {
      final Wrapper<Position> p =     assertCreatePosition(5, 6);

      assertMove(p, Direction.N, true, "(5|7)");
      assertMove(p, Direction.E, true, "(6|7)");
      assertMove(p, Direction.SE, true, "(7|6)");
      assertMove(p, Direction.W, true, "(6|6)");
    }
    {
      final Wrapper<Position> p = assertCreatePosition(0, 0);
      assertMove(p, Direction.NE, true, "(1|1)");
    }
    {
      final Wrapper<Position> p = assertCreatePosition(0, 0);
      assertMove(p, Direction.N, true, "(0|1)");
    }
    {
      final Wrapper<Position> p = assertCreatePosition(0, 0);
      assertMove(p, Direction.E, true, "(1|0)");
    }
  }

  @Test @Marking(points = 8) public void test_160_illegalMoves() {
    {
      final Wrapper<Position> p = assertCreatePosition(10, 10);

      // Top right (Northeast) position, impossible to move north- or eastbound
      assertMove(p, Direction.E, false, "(10|10)");
      assertMove(p, Direction.NE, false, "(10|10)");
      assertMove(p, Direction.N, false, "(10|10)");
      assertMove(p, Direction.NE, false, "(10|10)");

      assertMove(p, Direction.W, true, "(9|10)");
      assertMove(p, Direction.E, true, "(10|10)");

      assertMove(p, Direction.W, true, "(9|10)");
      assertMove(p, Direction.W, true, "(8|10)");
      assertMove(p, Direction.N, false, "(8|10)");
      assertMove(p, Direction.S, true, "(8|9)");
    }
  }

  /**
   * <p>Check both a move's success status and its new position value.</p>
   *
   * @param p Starting position
   * @param d Desired direction of movement
   * @param expectedStatus Move may either succeed or fail e.g. when staying on the grid's border.
   * @param expectedPosition Outcome e.g. »(3|7)«
   */
  private static void assertMove(final Wrapper<Position> p,
                                 final Direction d,
                                 final boolean expectedStatus,
                                 final String expectedPosition) {
    Assert.assertEquals(expectedStatus, p.invokeBoolean("move", P.n(d)));
    Assert.assertEquals(expectedPosition, p.invokeObject("toString"));
  }

  private static Wrapper<Position> assertCreatePosition(final int x, final int y) {

    final Wrapper<Position> p = new Wrapper<>(Position.class, P.n(x), P.n(y));
    Assert.assertEquals("(" + x + "|" + y + ")", p.invokeObject("toString"));
    return p;
  }
}
